@@ -891,7 +891,7 @@ int collect_user_ns(struct ns_id *ns, void *oarg)
 {
 	/*
 	 * User namespace is dumped before files to get uid and gid
-	 * mappings, which are used for convirting local id-s to
+	 * mappings, which are used for converting local id-s to
 	 * userns id-s (userns_uid(), userns_gid())
 	 */
 	if (dump_user_ns(root_item->pid->real, root_item->ids->user_ns_id))
@@ -1330,11 +1330,6 @@ static int usernsd(int sk)
 		unsc_msg_pid_fd(&um, &pid, &fd);
 		pr_debug("uns: daemon calls %p (%d, %d, %x)\n", call, pid, fd, flags);
 
-		if (fd < 0 && flags & UNS_FDOUT) {
-			pr_err("uns: bad flags/fd %p %d %x\n", call, fd, flags);
-			BUG();
-		}
-
 		/*
 		 * Caller has sent us bare address of the routine it
 		 * wants to call. Since the caller is fork()-ed from the
