@@ -92,6 +92,7 @@ void criu_set_manage_cgroups(bool manage);
 void criu_set_manage_cgroups_mode(enum criu_cg_mode mode);
 int criu_set_freeze_cgroup(const char *name);
 int criu_set_lsm_profile(const char *name);
+int criu_set_lsm_mount_context(const char *name);
 void criu_set_timeout(unsigned int timeout);
 void criu_set_auto_ext_mnt(bool val);
 void criu_set_ext_sharing(bool val);
@@ -111,6 +112,7 @@ int criu_set_pre_dump_mode(enum criu_pre_dump_mode mode);
 void criu_set_pidfd_store_sk(int sk);
 int criu_set_network_lock(enum criu_network_lock_method method);
 int criu_join_ns_add(const char *ns, const char *ns_file, const char *extra_opt);
+void criu_set_mntns_compat_mode(bool val);
 
 /*
  * The criu_notify_arg_t na argument is an opaque
@@ -160,6 +162,7 @@ int criu_get_orphan_pts_master_fd(void);
  */
 int criu_check(void);
 int criu_dump(void);
+int criu_pre_dump(void);
 int criu_restore(void);
 int criu_restore_child(void);
 
@@ -184,7 +187,7 @@ int criu_dump_iters(int (*more)(criu_predump_info pi));
  * As this library is just forwarding all tasks to an
  * independent (of this library) CRIU binary, the actual
  * version of the CRIU binary can be different then the
- * hardcoded values in the libary (version.h).
+ * hardcoded values in the library (version.h).
  * To be able to easily check the version of the CRIU binary
  * the function criu_get_version() returns the version
  * in the following format:
@@ -249,6 +252,7 @@ void criu_local_set_manage_cgroups(criu_opts *opts, bool manage);
 void criu_local_set_manage_cgroups_mode(criu_opts *opts, enum criu_cg_mode mode);
 int criu_local_set_freeze_cgroup(criu_opts *opts, const char *name);
 int criu_local_set_lsm_profile(criu_opts *opts, const char *name);
+int criu_local_set_lsm_mount_context(criu_opts *opts, const char *name);
 void criu_local_set_timeout(criu_opts *opts, unsigned int timeout);
 void criu_local_set_auto_ext_mnt(criu_opts *opts, bool val);
 void criu_local_set_ext_sharing(criu_opts *opts, bool val);
@@ -272,11 +276,13 @@ int criu_local_set_pre_dump_mode(criu_opts *opts, enum criu_pre_dump_mode mode);
 void criu_local_set_pidfd_store_sk(criu_opts *opts, int sk);
 int criu_local_set_network_lock(criu_opts *opts, enum criu_network_lock_method method);
 int criu_local_join_ns_add(criu_opts *opts, const char *ns, const char *ns_file, const char *extra_opt);
+void criu_local_set_mntns_compat_mode(criu_opts *opts, bool val);
 
 void criu_local_set_notify_cb(criu_opts *opts, int (*cb)(char *action, criu_notify_arg_t na));
 
 int criu_local_check(criu_opts *opts);
 int criu_local_dump(criu_opts *opts);
+int criu_local_pre_dump(criu_opts *opts);
 int criu_local_restore(criu_opts *opts);
 int criu_local_restore_child(criu_opts *opts);
 int criu_local_dump_iters(criu_opts *opts, int (*more)(criu_predump_info pi));
@@ -284,6 +290,35 @@ int criu_local_dump_iters(criu_opts *opts, int (*more)(criu_predump_info pi));
 int criu_local_get_version(criu_opts *opts);
 int criu_local_check_version(criu_opts *opts, int minimum);
 
+/*
+ * Feature checking allows the user to check if CRIU supports
+ * certain features. There are CRIU features which do not depend
+ * on the version of CRIU but on kernel features or architecture.
+ *
+ * One example is memory tracking. Memory tracking can be disabled
+ * in the kernel or there are architectures which do not support
+ * it (aarch64 for example). By using the feature check a libcriu
+ * user can easily query CRIU if a certain feature is available.
+ *
+ * The features which should be checked can be marked in the
+ * structure 'struct criu_feature_check'. Each structure member
+ * that is set to true will result in CRIU checking for the
+ * availability of that feature in the current combination of
+ * CRIU/kernel/architecture.
+ *
+ * Available features will be set to true when the function
+ * returns successfully. Missing features will be set to false.
+ */
+
+struct criu_feature_check {
+	bool mem_track;
+	bool lazy_pages;
+	bool pidfd_store;
+};
+
+int criu_feature_check(struct criu_feature_check *features, size_t size);
+int criu_local_feature_check(criu_opts *opts, struct criu_feature_check *features, size_t size);
+
 #ifdef __GNUG__
 }
 #endif
