@@ -1,3 +1,5 @@
+#!/bin/bash
+
 # Check lazy-pages
 set -e
 source `dirname $0`/criu-lib.sh
@@ -13,7 +15,7 @@ LAZY_MIGRATE_EXCLUDE="-x fifo_loop -x file_locks -x ptrace_sig -x overmount_file
 	       --lazy-migrate $LAZY_EXCLUDE $LAZY_MIGRATE_EXCLUDE || fail
 
 # During pre-dump + lazy-pages we leave VM_NOHUGEPAGE set
-LAZY_EXCLUDE="$LAZY_EXCLUDE -x maps02"
+LAZY_EXCLUDE="$LAZY_EXCLUDE -x maps02 -x maps09 -x maps10"
 
 # lazy restore from images with pre-dumps
 ./test/zdtm.py run --all --keep-going --report report --parallel 4 -f uns \
